#!/usr/bin/env python
"""
Activity graph suite
WSU CASAS Project
"""

# Import Python libraries
import sys
import re
import copy
import datetime
import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import pylab
import numpy
from matplotlib.pyplot import figure, show, rc, axes, bar
from matplotlib import cm, colors
from matplotlib.font_manager import FontProperties
from Numeric import *

# Global variable declaration
global_begin = "begin"
global_end = "end"


"""
Takes the given string in format of YYYY-MM-DD HH:MM:SS.ms and converts it
to a datetime.datetime() object.  The ms field is optional.
    
newVal - string to convert to a datetime.datetime() object.
"""
def get_datetime(newVal):
   stuff = re.split('\s+', newVal)
   date = re.split('-', stuff[0])
   time = re.split(':', stuff[1])
   sec = []
   if re.search('\.', time[2]) == None:
      sec.append(time[2])
      sec.append("0")
   else:
      sec = re.split('\.', time[2])
   dt = datetime.datetime(int(date[0]),
                          int(date[1]),
                          int(date[2]),
                          int(time[0]),
                          int(time[1]),
                          int(sec[0]),
                          int(sec[1]))
   return dt


"""
Define annotation bar position and content.
"""
class AnnBar:
   def __init__(self, name):
      self.name = copy.copy(name)
      self.left = []
      self.height = []
      self.width = []
      self.bottom = []
      return
    
   def __str__(self):
      mystr = "    %s\n" % self.name
      mystr += "      left = %s %s\n" % (str(len(self.left)),str(self.left))
      mystr += "      higt = %s %s\n" % (str(len(self.height)),str(self.height))
      mystr += "      widt = %s %s\n" % (str(len(self.width)),str(self.width))
      mystr += "      botm = %s %s\n" % (str(len(self.bottom)),str(self.bottom))
      return mystr
    
   def add_left(self, lnum):
      self.left.append(lnum)
      return
    
   def add_height(self, hnum):
      self.height.append(hnum)
      return
    
   def add_width(self, wnum):
      self.width.append(wnum)
      return
    
   def add_bottom(self, bnum):
      self.bottom.append(bnum)
      return


"""
Define day bar class variables and functions.  A day bar graphs all of the
activities for a single day.
"""
class DayBar:
   def __init__(self, date, annVals):
      self.date = copy.copy(date)
      self.bars = dict()
      for x in annVals:
         self.bars[x] = AnnBar(x)
      return
    
   def __str__(self):
      mystr = "%s\n" % str(self.date)
      for x in self.bars:
         mystr += "%s" % str(self.bars[x])
      return mystr
    
   def plot(self, ax, colors):
      for x in self.bars:
         ax.bar(left=self.bars[x].left,
                height=self.bars[x].height,
                width=self.bars[x].width,
                bottom=self.bars[x].bottom,
                linewidth=0,
                color=colors[x],
                label=str(x))
      return
    
   def build_bar(self, data):
      opnList = []
      impList = []
      dayTime = datetime.datetime.combine(self.date, datetime.time(0))
      dayLast = datetime.datetime.combine(self.date, datetime.time(23,59,59))
      # Gather "important points"
      for x in range(len(data)):
         if self.date == data[x].start.date() or self.date == data[x].end.date():
            impList.append(RawData(data[x].start, data[x].annot, global_begin))
            impList.append(RawData(data[x].end, data[x].annot, global_end))
      impList.sort()

      # Insert values into self.bars
      for x in range(len(impList)):
         if impList[x].state == global_begin:
            opnList.append(impList[x].annot)
         else:
            try:
               opnList.remove(impList[x].annot)
            except:
               print impList[x].annot
               print opnList
                
         # Add left values
         if impList[x].dt < dayTime:
            for y in self.bars:
               self.bars[y].add_left(0)
         else:
            diff = impList[x].dt - dayTime
            for y in self.bars:
               self.bars[y].add_left(diff.seconds)
            
         # Add width values
         if (x+1) >= len(impList):
            nextDT = dayLast
         else:
            nextDT = impList[x+1].dt
         if nextDT > (dayTime + datetime.timedelta(1)):
            diff = (dayTime + datetime.timedelta(1)) - impList[x].dt
            for y in self.bars:
               self.bars[y].add_width(diff.seconds)
         else:
            diff = nextDT - impList[x].dt
            for y in self.bars:
               self.bars[y].add_width(diff.seconds)
            
         # Add height values
         for y in self.bars:
            if impList[x].annot == y or y in opnList:
               if len(opnList) != 0:
                  self.bars[y].add_height(1.0/float(len(opnList)))
               else:
                  self.bars[y].add_height(0.0)
            else:
               self.bars[y].add_height(0.0)
                
         # bottom values AFTER this for loop.
      # Calculate bottom values
      kys = self.bars.keys()
      for x in range(len(self.bars[kys[0]].left)):
         bot = 0.0
         for y in self.bars:
            self.bars[y].add_bottom(bot)
            bot += self.bars[y].height[x]
      return


"""
Define day RawData class variables and functions.  RawData stores date and
time information for a set of activities.
"""
class RawData:
   def __init__(self, dt, annotation, state):
      self.dt = copy.copy(dt)
      self.annot = copy.copy(annotation)
      self.state = copy.copy(state)
      return
    
   def __cmp__(self, other):
      val = 0
      if self.dt < other.dt:
         val = -1
      elif self.dt > other.dt:
         val = 1
      return val
    
   def __str__(self):
      mystr = "%s\tx\tx\t%s\t%s" % (str(self.dt),str(self.annot),str(self.state))
      return mystr


"""
Define day AnnData class variables and functions.  AnnData stores the
start time, end time, and number of other concurrent activities for
each activity being tracked.
"""
class AnnData:
   def __init__(self, start, end, annotation, concurrent):
      self.start = copy.copy(start)
      self.end = copy.copy(end)
      self.annot = copy.copy(annotation)
      self.conc = copy.copy(concurrent)
      return
    
   def __cmp__(self, other):
      val = 0
      if self.start < other.start:
         val = -1
      elif self.start > other.start:
         val = 1
      return val
    
   def __str__(self):
      mystr = "%s\t%s\t%s\t%s" % (str(self.start),
                                  str(self.end),
                                  str(self.conc),
                                  str(self.annot))
      return mystr


"""
Define Viz class variables and function.  Process data files and
generate graphs.
"""
class Viz:
   def __init__(self, files, toPng):
      self.files = files
      self.dates = []
      self.toPng = toPng
      self.data = []
      self.a = []
      self.adata = []
      self.annotations = []
      self.daybars = []
      self.counts = dict([('Bathing', 0), ('Bed_to_Toilet', 0), ('Cook', 0), ('Eat', 0), ('Enter_Home', 0), ('Leave_Home', 0), ('Other', 0), ('Personal_Hygiene', 0), ('Relax', 0), ('Sleep', 0), ('Take_Medicine', 0)])
      self.times = dict()
      self.filtered = []
      self.colorList = []
      return
 
   # Extract start and end times for activities, store in separate file
   def summarize_dates(self):
      filename = self.files[0]
      input = open(filename)
      for line in input:
         x = str(str(line).strip()).split()
         if len(x) > 5:
	    self.filtered.append(x)
      input.close()
      return

   # Extract dates from input file
   def prepare_files(self):
      filename = self.files[0]
      input = open(filename)
      for line in input:
         x = str(str(line).strip()).split()
	 lDate = x[0]
	 if lDate not in self.dates:
	    self.dates.append(lDate)
      self.dates.sort()
      numDates = len(self.dates)
      input.close()
      self.a = zeros([numDates, 24], Int)
      return
    
   def load_files(self):
      previous = None
      pieOpenList = []
      pieOpenList.append('Other')
      pie2OpenList = []
      startTimes = []
      firstLine = 1
      filename = self.files[0]
      errors = ""
      print "Reading file: %s" % filename

      input = open(filename)

      # Loop through the given file.
      for line in input:
         event = str(str(line).strip()).split()
         lDate = event[0]
         lTime = event[1]
	 gTime = get_datetime(str(lDate)+" "+str(lTime))
	 tvalue = int(str(gTime.hour))
	 tdate = self.dates.index(lDate)
	 self.a[tdate,tvalue] = self.a[tdate,tvalue] + 1
	 if firstLine == 1:
	    previous = gTime
	    diff = gTime - gTime
	    firstLine = 0
	    self.times['Bathing'] = diff
	    self.times['Bed_to_Toilet'] = diff
	    self.times['Cook'] = diff
	    self.times['Eat'] = diff
	    self.times['Enter_Home'] = diff
	    self.times['Leave_Home'] = diff
	    self.times['Other'] = diff
	    self.times['Personal_Hygiene'] = diff
	    self.times['Relax'] = diff
	    self.times['Sleep'] = diff
	    self.times['Take_Medicine'] = diff
                
         # If line contains annotated info, deal with it.
         if len(event) > 5:
	    activity = event[4]
	    state = event[5]
            if state == global_begin:     # Activity beginning
               pieOpenList.append(event[4])
	       pie2OpenList.append(event[4])
	       startTimes.append(gTime)
            else:                         # Activity ending
	       i = pie2OpenList.index(event[4])
	       j = startTimes[i]
	       for x in self.times:
	          if x in pie2OpenList:
		     i = pie2OpenList.index(x)
		     sTime = startTimes[i]
		     num = self.times[x]
		     diff = gTime - sTime
		     num = num + diff
		     del self.times[x]
		     self.times[x] = num
	       pieOpenList.remove(event[4])
	       pie2OpenList.remove(event[4])
	       startTimes.remove(j)
               previous = gTime
	 elif len(pie2OpenList) == 0:         # Add time to Other activity
	    diff = gTime - previous
	    num = self.times['Other']
	    num = num + diff
	    del self.times['Other']
	    self.times['Other'] = num
	    previous = gTime
         for x in self.counts:
	    if x in pieOpenList:
	       num = self.counts[x]
	       num = num + 1
	       del self.counts[x]
	       self.counts[x] = num
      self.adata = ravel(self.a)
      input.close()
      print errors
      return
    
   def load_activity_files(self, userData):
      opnList = dict()
      firstTime = None
      lDate = None
      lTime = None
      errors = ""

      # Start with annotations that are common to all files
      self.annotations.append("Bathing")
      self.annotations.append("Bed_to_Toilet")
      self.annotations.append("Cook")
      self.annotations.append("Eat")
      self.annotations.append("Enter_Home")
      self.annotations.append("Leave_Home")
      self.annotations.append("Other")
      self.annotations.append("Personal_Hygiene")
      self.annotations.append("Relax")
      self.annotations.append("Sleep")
      self.annotations.append("Take_Medicine")

      # Loop through the given file.
      for x in self.filtered:
         #x = str(str(line).strip()).split()
	 if firstTime == None:
	    firstTime = get_datetime(str(x[0])+" "+str(x[1]))
	 lDate = x[0]
	 lTime = x[1]
         # If line contains annotated info, deal with it.
         if len(x) > 5:
            dt = get_datetime(str(x[0])+" "+str(x[1]))
            if x[4] in opnList:
               dtStart = opnList[x[4]]
               if x[5] == global_end:
                  self.data.append(AnnData(dtStart,dt,x[4],len(opnList)))
                  del opnList[x[4]]
               if x[4] not in self.annotations:
                  self.annotations.append(x[4])
            else:
               if x[5] == global_end:
                  print "*"
                  errors += "  ERROR: *NO START* %s %s\n" % (str(dt), x[4])
                  if x[4] not in self.annotations:
                     self.annotations.append(x[4])
               else:
                  opnList[x[4]] = dt
      dt = get_datetime(str(lDate)+" "+str(lTime))
      for x in opnList:
         print "*"
         errors += "  ERROR: *NO END* %s %s\n" % (str(opnList[x]), str(x))
         if x not in self.annotations:
            self.annotations.append(x)
      print errors
      return

   def build_day_bars(self):
      tmp = self.get_dates()
      tmp.sort()
      for x in tmp:
         self.daybars.append(DayBar(x, self.annotations))
      for x in self.daybars:
         x.build_bar(self.data)
      return
    
   def get_dates(self):
      tmp = []
      for x in self.data:
         if x.start.date() not in tmp:
            tmp.append(x.start.date())
         if x.end.date() not in tmp:
            tmp.append(x.end.date())
      return tmp
    
   def get_important_points(self):
      imp = []
      day = datetime.datetime.combine(self.data[0].start.date(),datetime.time(0))
      for x in range(len(self.data)):
         diff = self.data[x].start - day
         imp.append(diff.seconds)
         if (x+1) < len(self.data):
            if self.data[x].end > self.data[x+1].start:
               diff = self.data[x].end - day
               imp.append(diff.seconds)
      return imp
    
   def build_colors(self):
      map = cm.gist_rainbow(pylab.linspace(0,1,len(self.annotations)))
      return map

   def show_pie_chart(self, toPng, fig):
      ax = plt.axes([0.0, 0.0, 0.25, 0.4])
      figtitle = 'Sensor frequency distribution'
      fig.text(0.12, 0.38, figtitle, horizontalalignment='center',
               fontproperties=FontProperties(size=8))
      names = []
      values = []
      for x in self.annotations:
         y = self.counts[x]
	 names.append(x)
	 values.append(y)

      # Make the pie figure
      pdata = plt.pie(values, labels=None, colors=self.colorList)
      return

   def show_time_pie_chart(self, toPng, fig):
      ax = plt.axes([0.3, 0.00, 0.25, 0.4])
      figtitle = 'Activity time distribution'
      fig.text(0.42, 0.38, figtitle, horizontalalignment='center',
               fontproperties=FontProperties(size=8))
      names = []
      values = []
      for x in self.annotations:
         y = self.times[x]
	 names.append(x)
	 values.append(y.seconds)

      # Make the pie figure
      pdata = plt.pie(values, labels=None, colors=self.colorList)
      plt.legend(pdata[0], names, loc=(1.5,0.05), prop=FontProperties(size=8))
      return

   def setup_graphs(self, toPng):
      print "Building Graph... This could take a while..."

      # Figure properties
      figprops = dict(figsize=(10., 10. / 1.618), dpi=128, facecolor='w')

      # Subplot properties
      adjustprops = dict(left=0.06, bottom=0.50, right=0.50, top=0.94,
                         wspace=0.2, hspace=0.01)
      # New figure
      fig = pylab.figure(**figprops)
      # Tunes the subplot layout
      fig.subplots_adjust(**adjustprops)
        
      # Build annotation colors
      self.colorList = self.build_colors()
      return fig
    
   def show_activity_graph(self, toPng, fig):
      figtitle = 'Activity occurrences'
      fig.text(0.25, 0.95, figtitle, horizontalalignment='center',
               fontproperties=FontProperties(size=8))

      colrDict = dict()
      for x in range(len(self.annotations)):
         colrDict[self.annotations[x]] = self.colorList[x]
      ticks = [0,3,6,9,12,15,18,21,24]
      ticklabels = ["12am","3am","6am","9am","12pm","3pm","6pm","9pm","12am"]
      for x in range(len(ticks)):
         ticks[x] = float(ticks[x]) * 60.0 * 60.0
        
      # If only 1 day, simple, otherwise loop.
      if len(self.daybars) < 2:
         ax = fig.add_subplot(1,1,1)
         pylab.setp(ax.get_yticklabels(), visible=False)
         ax.set_ylabel(str(self.daybars[0].date), fontsize=10)
         ax.set_xticks(ticks)
         ax.set_xticklabels(ticklabels, size=6, visible=True)
         self.daybars[0].plot(ax, colrDict)
      else:
         ax = fig.add_subplot(len(self.daybars),1,1, frame_on=False)
         pylab.setp(ax.get_yticklabels(), visible=False)
         pylab.setp(ax.get_xticklabels(), visible=False)
         ax.set_ylabel(self.daybars[0].date,
                       fontsize=6, rotation='horizontal')
         self.daybars[0].plot(ax, colrDict)
         for x in range(len(self.daybars)-1):
            bx = fig.add_subplot(len(self.daybars),1,x+2,
                                 sharex=ax, sharey=ax, frame_on=False)
            pylab.setp(bx.get_xticklabels(), visible=False)
            pylab.setp(bx.get_yticklabels(), visible=False)
            bx.set_ylabel(self.daybars[x+1].date,
                          fontsize=6, rotation='horizontal')
            self.daybars[x+1].plot(bx, colrDict)
         bx.set_yticks([0.0])
         bx.set_xticks(ticks)
         bx.set_xticklabels(ticklabels, size=6, visible=True)
      return

   def show_density_chart(self, toPng, fig):
      figtitle = 'Activity density'
      fig.text(0.8, 0.95, figtitle, horizontalalignment='center',
               fontproperties=FontProperties(size=8))

      # Choose a color map for the figure
      cmap = cm.spectral

      # Add axes with left/bottom/width/height fraction of figure width/height
      a = fig.add_axes([0.6, 0.56, 0.39, 0.38])    # Add plot area
      cax = fig.add_axes([0.6, 0.50, 0.39, 0.02]) # Add colorbar
      for tick in cax.xaxis.get_major_ticks():
         tick.label1.set_fontsize(6)
      for tick in a.xaxis.get_major_ticks():
         tick.label1.set_fontsize(6)
      for tick in a.yaxis.get_major_ticks():
         tick.label1.set_fontsize(6)

      # Manually find the min and max of all colors for
      # use in setting the color scale.
      vmin = 1e40
      vmax = -1e40
      vmin = min(vmin, numpy.amin(self.adata))
      vmax = max(vmax, numpy.amax(self.adata))
      ticks = []
      showdates = []
      onequarter = int(len(self.dates)/4)
      halfway = int(len(self.dates)/2)
      threequarters = 3*int(len(self.dates)/4)
      ticks.insert(0,0)
      ticks.insert(0,onequarter)
      ticks.insert(0,halfway)
      ticks.insert(0,threequarters)
      ticks.insert(0,len(self.dates)-1)
      showdates.insert(0,self.dates[0])
      showdates.insert(0,self.dates[onequarter])
      showdates.insert(0,self.dates[halfway])
      showdates.insert(0,self.dates[threequarters])
      showdates.insert(0,self.dates[len(self.dates)-1])
      a.set_yticks(ticks)
      a.set_yticklabels(showdates)
      image = a.imshow(self.a, cmap=cmap, aspect='auto', interpolation=None,
                       alpha=1)

      norm = colors.Normalize(vmin=vmin, vmax=vmax)
      image.set_norm(norm)

      # The colorbar is also based on this master image.
      fig.colorbar(image, cax, orientation='horizontal')
      return

   def show_graphs(self, toPng, fig):
      if toPng == True:
         plt.savefig("activitychart.png")
      plt.show()
      return 


if __name__ == "__main__":
   files = []
   toPng = True
   if len(sys.argv[1:]) < 1:
      print "You need to specify a data file."
      print "annoViz.py filename"
      sys.exit(0)
    
   for x in sys.argv[1:]:
      files.append(x)
   viz = Viz(files, toPng)
   userData = viz.summarize_dates()
   viz.prepare_files()
   viz.load_files()
   viz.load_activity_files(userData)
   viz.build_day_bars()
   fig = viz.setup_graphs(toPng)
   viz.show_activity_graph(toPng, fig)
   viz.show_density_chart(toPng, fig)
   viz.show_pie_chart(toPng, fig)
   viz.show_time_pie_chart(toPng, fig)
   viz.show_graphs(toPng, fig)
